/*
 * Copyright (C) 2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include <ortp/ortp.h>
#include <vorbis/vorbisfile.h>
#include "ovsend.cmdline.h"


#define CHUNK_SIZE 40960

static RtpSession *session;
static char *header_buf;
static int header_len;
static struct gengetopt_args_info conf;

enum {END_OF_FILE, BEGIN_OF_FILE, OV_HEADER};

static void send_ctrlmsg(char *buf, int len, int ts)
{
	rtp_session_set_payload_type(session, PAYLOAD_OTHER);
	rtp_session_send_with_ts(session, buf, len, ts);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);
}

static void clog(int ll, char* fmt,...)
{
	va_list argp;

	/* ignore log message if loglevel is not high enough */
	if (ll < conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

static void send_notifier(int code, int ts)
{
	char buf[2];

	clog(DEBUG, "sending notifier #%d\n", code);
	if (code == OV_HEADER) {
		*header_buf = OV_HEADER;
		return send_ctrlmsg(header_buf, header_len, ts);
	}
	buf[0] = code;
	buf[1] = '\0';
	send_ctrlmsg(buf, 2, ts);
}

static int save_headers(int in, char *first_buf)
{
	int ret, len;
	unsigned int serial;
	char *buf;

	ogg_page page;
        ogg_packet packet;
	vorbis_comment vc;
	vorbis_info vi;
	ogg_stream_state *stream_in = malloc(sizeof(ogg_stream_state));
	ogg_stream_state *stream_out = malloc(sizeof(ogg_stream_state));
	ogg_sync_state *sync_in = malloc(sizeof(ogg_sync_state));

	ogg_sync_init(sync_in);
	buf = ogg_sync_buffer(sync_in, CHUNK_SIZE);
	ret = -1;
	len = read(in, first_buf, CHUNK_SIZE);
	if (len <= 0) {
		clog(len? ERROR : NOTICE, "read returned %d (%s)\n", len,
			len < 0? strerror(errno) : "success");
		goto out;
	}
	memcpy(buf, first_buf, len);
	ogg_sync_wrote(sync_in, len);
	if (ogg_sync_pageout(sync_in, &page) <= 0) {
		clog(ERROR, "%s: no ogg file!\n", __func__);
		goto out;
	}
	serial = ogg_page_serialno(&page);
	ogg_stream_init(stream_in, serial);
	ogg_stream_init(stream_out, serial);
	ret = ogg_stream_pagein(stream_in, &page);
	if (ret < 0) {
		clog(ERROR, "%s: error in first page\n", __func__);
		goto out;
	}
	ret = ogg_stream_packetout(stream_in, &packet);
	if (ret != 1) {
		clog(ERROR, "%s: error in first packet\n", __func__);
		goto out;
	}
	vorbis_info_init(&vi);
	vorbis_comment_init(&vc);
	if (vorbis_synthesis_headerin(&vi, &vc, &packet) < 0) {
		clog(ERROR, "%s: error in primary header: not vorbis?\n",
			__func__);
		goto out;
	} else
		clog(INFO, "channels: %i, rate: %li\n", vi.channels, vi.rate);
	ogg_stream_packetin(stream_out, &packet);

	ret = ogg_sync_pageout(sync_in, &page);
	ogg_stream_pagein(stream_in, &page);
	ogg_stream_packetout(stream_in, &packet);
	ogg_stream_packetin(stream_out, &packet);

	ret = ogg_sync_pageout(sync_in, &page);
	ogg_stream_pagein(stream_in, &page);
	ogg_stream_packetout(stream_in, &packet);
	ogg_stream_packetin(stream_out, &packet);

	header_len = 1; /* first byte is reserved for notifier code */
	while (ogg_stream_flush(stream_out, &page)) {
		ssize_t tmp = page.body_len + page.header_len;
		header_buf = realloc(header_buf, header_len + tmp);
		memcpy(header_buf + header_len, page.header, page.header_len);
		memcpy(header_buf + header_len + page.header_len, page.body,
			page.body_len);
		header_len += tmp;
	}
	*header_buf = OV_HEADER;
	ret = len;
out:
	clog(DEBUG, "%s: returning %d, header_len = %d\n", __func__, ret,
		header_len);
	vorbis_info_clear(&vi);
        vorbis_comment_clear(&vc);
	ogg_stream_destroy(stream_in);
	ogg_stream_destroy(stream_out);
	ogg_sync_destroy(sync_in);
	return ret;
}

static int do_io(int in, RtpSession *session)
{
	int sent = 0, ret, bytes, count = 0;
	static int ts;
	char buf[CHUNK_SIZE];
	struct timeval lhst; /* last header send time */
	long unsigned ms;

	ret = save_headers(in, buf);
	if (ret < 0)
		return ret;
	clog(DEBUG, "initial buffer: %d bytes\n", ret);
	ts = rtp_session_get_current_send_ts(session);
	send_notifier(BEGIN_OF_FILE, ts);
//	usleep(1000 * 200); /* testing */
	gettimeofday(&lhst, NULL);
	while (ret > 0) {
		struct timeval now, tmp;
		int hia = conf.header_interval_arg;

		count++;
		clog(DEBUG, "sending data (%d)\n", ret);
		ts = rtp_session_get_current_send_ts(session);
		bytes = rtp_session_send_with_ts(session, buf, ret, ts);
		sent += ret;
		gettimeofday(&now, NULL);
		tv_diff(&now, &lhst, &tmp);
		ms = tv2ms(&tmp);
		if (ms > hia * 1000 || count < hia) {
			if (ms > 2000 * hia)
				count = 0;
			clog(INFO, "sending headers (%d)\n", header_len);
			ts = rtp_session_get_current_send_ts(session);
			send_notifier(OV_HEADER, ts);
			lhst = now;
		}
//		usleep(1000 * 100); /* testing */
		ret = read(in, buf, CHUNK_SIZE);
	}
	ts = rtp_session_get_current_send_ts(session);
	send_notifier(END_OF_FILE, ts);
	if (!ret)
		usleep(1000 * 1000); /* testing */
	clog(DEBUG, "%s: sent %d bytes\n", __func__, sent);
	return ret;
}

int main(int argc, char *argv[])
{
	int i, ret;
	int infile;

	cmdline_parser(argc, argv, &conf);
	clog(NOTICE, "oggsend startup\n");
	if (!conf.inputs_num) {
		cmdline_parser_print_help();
		exit(EXIT_FAILURE);
	}
	ortp_init();
	ortp_scheduler_init();
	session = rtp_session_new(RTP_SESSION_SENDONLY);
	rtp_session_set_scheduling_mode(session, 1);
	rtp_session_set_blocking_mode(session, 1);
	rtp_session_set_remote_addr(session, conf.hostname_arg,
		conf.port_arg);
	rtp_session_set_payload_type(session, PAYLOAD_AUDIO_CONTINUOUS);
repeat:
	for (i = 0; i < conf.inputs_num; i++) {
		char *arg = conf.inputs[i];
		clog(INFO, "************* Next: %s ***********\n", arg);
		if (strcmp(arg, "-")) {
			infile = open(arg, O_RDONLY);
			ret = -1;
			if (!infile) {
				clog(ERROR, "can not open file\n");
				goto out;
			}
		} else
			infile = STDIN_FILENO;
		ret = do_io(infile, session);
		if (ret < 0)
			goto out;
		if (infile != STDIN_FILENO)
			close(infile);
	}
	//sleep(5); /* testing */
	if (conf.repeat_given)
		goto repeat;
out:
	rtp_session_destroy(session);
	ortp_exit();
//	ortp_global_stats_display();
	return ret;
}
